# TODO: these could use a bit of better error handling
# - style the NameErrors better
# - raise error if there's no __closure__ ?


def replace_closure_variables(func, **kwargs):
    variable_names = func.__code__.co_freevars
    cells = func.__closure__
    for k, v in kwargs.items():
        try:
            pos = variable_names.index(k)
        except ValueError:
            raise NameError(k)
        cells[pos].cell_contents = v


def get_closure_variable(func, name):
    variable_names = func.__code__.co_freevars
    cells = func.__closure__
    for pos, var_name in enumerate(variable_names):
        if var_name == name:
            return cells[pos].cell_contents
    raise NameError(name)




# Inspiration:
# - https://github.com/rfk/withhacks
# - https://code.google.com/archive/p/ouspg/wikis/AnonymousBlocksInPython.wiki
# - https://billmill.org/multi_line_lambdas.html
import functools
import sys
import threading


class StopSwitch(Exception):
    pass


class SkipCase(Exception):
    pass


_switch_lock = threading.Lock()

class case:
    def __init__(self, b, context):
        self.b = b
        self._context = context
        self._trace_func = None
        self._orig_sys_trace = None

    def __enter__(self):
        if not self:
            with _switch_lock:
                orig_trace = self._orig_sys_trace = sys.gettrace()
                if orig_trace is None:
                    self._trace_func = sys.settrace(lambda *a, **kw: None)
                frame = sys._getframe(1)
                frame.f_trace = self._trace_func = self.skip

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is SkipCase:
            if self._trace_func:
                with _switch_lock:
                    sys.settrace(self._orig_sys_trace)

            return True
        elif _is_break_exception(exc_type, exc_value):
            raise StopSwitch

    def skip(self, frame, event, arg):
        raise SkipCase

    def __bool__(self):
        if self._context.found:
            return True
        self._context.found = self._context.a == self.b
        return self._context.found


class switch:
    def __init__(self, a):
        self.a = a
        self.found = False

    def __iter__(self):
        yield functools.partial(case, context=self)

    def __enter__(self):
        return functools.partial(case, context=self)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is StopSwitch:
            return True
        elif _is_break_exception(exc_type, exc_value):
            return True


def _is_break_exception(exc_type, exc_value):
    return exc_type is NameError and \
        'break' in str(exc_value).lower()


def main():
    print('\nTest 1, for/if:')
    for case in switch('b'):
        if case('a'):
            print('this is not printed')
        if case('b'):
            print('this is printed')
        if case('c'):
            print('fallthrough is also supported')
            break
        if case('d'):
            print('this is not printed either')
        print('default value at bottom, but this is not reached')

    print('\nTest 2, for/with:')
    for case in switch('b'):
        with case('a'):
            print('this is not printed')
        with case('b'):
            print('this is printed')
        with case('c'):
            print('fallthrough is also supported')
            break
        with case('d'):
            print('this is not printed either')
        print('default value at bottom, but this is not reached')


    print('\nTest 3, with/with:')
    with switch('b') as case:
        with case('a'):
            print('this is not printed')
        with case('b'):
            print('this is printed')
        with case('c'):
            print('fallthrough is also supported')
            BREAK  # NOTE: We have to yell here
        with case('d'):
            print('this is not printed either')
        print('default value at bottom, but this is not reached')


if  __name__ == '__main__':
    main()



from os import path
from setuptools import setup, find_packages


here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='stupidity',
    version='0.0.1',
    url='https://github.com/koirikivi/stupidity.git',
    author='Rainer Koirikivi',
    author_email='rainer@koirikivi.fi',
    description='Collection of stupid python modules',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[],
    extras_require={
        'testing': [
            'pytest',
            'pytest-watch',
        ]
    },
    keywords=['stupid', 'stupidity', 'idiocy'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



# TODO
import ast
import inspect
import textwrap


def nostalgia(func):
    # Is there way to parse ast directly from func.__code__ without
    # inspect.getsource()??
    source = textwrap.dedent(inspect.getsource(func))
    source = source.lstrip('@nostalgia')  # err...
    #print(source)
    tree = ast.parse(source)
    tree = NostalgiaTransformer().visit(tree)
    tree = ast.fix_missing_locations(tree)
    print(ast.dump(tree).replace('(', '\n('))
    compiled_module = compile(tree, '<nostalgia>', 'exec')
    locals_ = {}
    exec(compiled_module, globals(), locals_)
    print(compiled_module, locals_)
    return locals_[func.__name__]


class NostalgiaTransformer(ast.NodeTransformer):
    def visit_BinOp(self, node):
        # Replace x / y with x // y if x and y are integers
        if isinstance(node.op, ast.Div) and \
                isinstance(node.left, ast.Num) and \
                isinstance(node.right, ast.Num) and \
                isinstance(node.left.n, int) and \
                isinstance(node.right.n, int):
            node.op = ast.FloorDiv()
        return node


def main():
    @nostalgia
    def foo():
        import sys
        from . import foobar
        x = 3 / 2
        return x

    assert foo() == 1

if __name__ == '__main__':
    main()
